# -*- coding: utf-8 -*-
"""
Helper functions for handling JSON data from Ostriz.
"""

from __future__ import absolute_import, unicode_literals

import datetime
import io
import json
import logging
import os
from collections import OrderedDict

import requests
import six

from dump2polarion import xunit_exporter
from dump2polarion.exceptions import Dump2PolarionException, NothingToDoException

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


IGNORED_PARAMS = {"browserVersion", "browserPlatform", "browserName"}


def _get_json(location):
    """Reads JSON data from file or URL."""
    location = os.path.expanduser(location)
    try:
        if os.path.isfile(location):
            with io.open(location, encoding="utf-8") as json_data:
                return json.load(json_data, object_pairs_hook=OrderedDict).get("tests")
        elif "http" in location:
            json_data = requests.get(location)
            if not json_data:
                raise Dump2PolarionException("Failed to download")
            return json.loads(json_data.text, object_pairs_hook=OrderedDict).get("tests")
        else:
            raise Dump2PolarionException("Invalid location")
    except Exception as err:
        raise Dump2PolarionException("Failed to parse JSON from {}: {}".format(location, err))


def _get_testrun_id(version):
    """Gets testrun id out of the appliance_version file."""
    try:
        build_base = version.strip().split("-")[0].split("_")[0].replace(".", "_")
        zval = int(build_base.split("_")[3])
    except Exception:
        # not in expected format
        raise Dump2PolarionException("Cannot find testrun id")
    if zval < 10:
        pad_build = build_base[-1].zfill(2)
        return build_base[:-1] + pad_build
    return build_base


def _calculate_duration(start_time, finish_time):
    """Calculates how long it took to execute the testcase."""
    if not (start_time and finish_time):
        return 0
    start = datetime.datetime.fromtimestamp(start_time)
    finish = datetime.datetime.fromtimestamp(finish_time)
    duration = finish - start

    decimals = float(("0." + str(duration.microseconds)))
    return duration.seconds + decimals


def _get_testname(test_path):
    """Gets test name out of full test path."""
    path_end = test_path.find(".py/")
    if path_end:
        offset = path_end + 4
        return test_path[offset:]
    return None


def _filter_parameters(parameters):
    """Filters the ignored parameters out."""
    if not parameters:
        return None
    return OrderedDict(
        (param, value) for param, value in six.iteritems(parameters) if param not in IGNORED_PARAMS
    )


def _append_record(test_data, results, test_path):
    """Adds data of single testcase results to results database."""
    statuses = test_data.get("statuses")
    jenkins_data = test_data.get("jenkins") or {}

    data = [
        ("title", test_data.get("test_name") or _get_testname(test_path)),
        ("verdict", statuses.get("overall")),
        ("source", test_data.get("source")),
        ("job_name", jenkins_data.get("job_name")),
        ("run", jenkins_data.get("build_number")),
        ("params", _filter_parameters(test_data.get("params"))),
        (
            "time",
            _calculate_duration(test_data.get("start_time"), test_data.get("finish_time")) or 0,
        ),
    ]
    test_id = test_data.get("polarion")
    if test_id:
        if isinstance(test_id, list):
            test_id = test_id[0]
        data.append(("test_id", test_id))

    results.append(OrderedDict(data))


def _comp_finish_time(test_data, last_finish_time):
    curr_finish_time = test_data.get("finish_time") or 0
    if curr_finish_time > last_finish_time[0]:
        last_finish_time[0] = curr_finish_time


def _parse_ostriz(ostriz_data):
    """Reads the content of the input JSON and returns testcases results."""
    if not ostriz_data:
        raise NothingToDoException("No data to import")

    results = []
    found_build = None
    last_finish_time = [0]
    for test_path, test_data in six.iteritems(ostriz_data):
        curr_build = test_data.get("build")
        if not curr_build:
            continue

        # set `found_build` from first record where it's present
        if not found_build:
            found_build = curr_build

        # make sure we are collecting data for the same build
        if found_build != curr_build:
            continue

        if not test_data.get("statuses"):
            continue

        _append_record(test_data, results, test_path)
        _comp_finish_time(test_data, last_finish_time)

    if last_finish_time[0]:
        logger.info("Last result finished at %s", last_finish_time[0])

    testrun_id = _get_testrun_id(found_build)
    return xunit_exporter.ImportedData(results=results, testrun=testrun_id)


# pylint: disable=unused-argument
def import_ostriz(location, **kwargs):
    """Reads Ostriz's data and returns imported data."""
    ostriz_data = _get_json(location)
    return _parse_ostriz(ostriz_data)



# -*- coding: utf-8 -*-
"""
Configuration loading.
"""

from __future__ import absolute_import, unicode_literals

import io
import logging
import os

import six
import yaml

from dump2polarion.exceptions import Dump2PolarionException

DEFAULT_USER_CONF = "~/.config/dump2polarion.yaml"
URLS = {
    "testcase_taget": "import/testcase",
    "xunit_target": "import/xunit",
    "requirement_target": "import/requirement",
    "testcase_queue": "import/testcase-queue",
    "xunit_queue": "import/xunit-queue",
    "requirement_queue": "import/requirement-queue",
    "testcase_log": "import/testcase-log",
    "xunit_log": "import/xunit-log",
    "requirement_log": "import/requirement-log",
    "auth_url": "j_security_check",
}

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def _check_config(config):
    missing = []
    for key in six.iterkeys(URLS):
        if not config.get(key):
            missing.append(key)

    # the 'auth_url' is allowed to be empty for now
    # TODO: can be removed once basic auth is discontinued on prod
    if not config.get("auth_url") and "auth_url" in config:
        missing.remove("auth_url")

    if missing:
        raise Dump2PolarionException(
            "Failed to find following keys in config file: {}\n"
            "Please see https://mojo.redhat.com/docs/DOC-1098563#config".format(", ".join(missing))
        )


def _guess_base_url(config):
    for key, value in six.iteritems(URLS):
        if config.get(key):
            return config[key][: -len(value)]
    return None


def _populate_urls(config):
    base_url = config.get("polarion_url")
    if not base_url:
        base_url = _guess_base_url(config)
    if not base_url:
        return

    base_url = base_url.rstrip("/")
    for key, url in six.iteritems(URLS):
        if key not in config:
            config[key] = "{}/{}".format(base_url, url)


def _set_project_id(config):
    if config.get("polarion-project-id"):
        return
    # use legacy configuration if available
    xunit_project = config.get("xunit_import_properties", {}).get("polarion-project-id")
    if xunit_project:
        config["polarion-project-id"] = xunit_project
        logger.warning(
            'Loading the "polarion-project-id" from legacy configuration "xunit_import_properties"'
            " instead of from top level"
        )
    else:
        raise Dump2PolarionException('The "polarion-project-id" key is missing in the config file')


def get_config(config_file=None):
    """Loads config file and returns its content."""
    default_conf = os.path.join(os.path.dirname(os.path.abspath(__file__)), "dump2polarion.yaml")
    user_conf = os.path.expanduser(config_file or DEFAULT_USER_CONF)

    try:
        with open(user_conf):
            pass
    except EnvironmentError:
        user_conf = None
        if config_file:
            raise Dump2PolarionException("Cannot open config file {}".format(config_file))

    with io.open(default_conf, encoding="utf-8") as input_file:
        config_settings = yaml.load(input_file)
    logger.debug("Default config loaded from %s", default_conf)

    if user_conf:
        with io.open(user_conf, encoding="utf-8") as input_file:
            config_settings_user = yaml.load(input_file)
        logger.info("Config loaded from %s", user_conf)

        # merge default and user configuration
        try:
            config_settings.update(config_settings_user)
        except ValueError as err:
            raise Dump2PolarionException(
                "Failed to load the '{}' config file: {}".format(user_conf, err)
            )

    _populate_urls(config_settings)
    _set_project_id(config_settings)
    _check_config(config_settings)

    return config_settings



# -*- coding: utf-8 -*-
"""
Submit data to the Polarion Importer.
"""

from __future__ import absolute_import, unicode_literals

import logging
import os

from dump2polarion import configuration, properties, utils
from dump2polarion.exceptions import Dump2PolarionException
from dump2polarion.verify import verify_submit

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def _get_xml_root(xml_root, xml_str, xml_file):
    if xml_root is not None:
        return xml_root
    if xml_str:
        return utils.get_xml_root_from_str(xml_str)
    if xml_file:
        return utils.get_xml_root(xml_file)
    raise Dump2PolarionException("Failed to submit to Polarion - no data supplied")


def _get_submit_target(xml_root, config):
    if xml_root.tag == "testcases":
        target = config.get("testcase_taget")
    elif xml_root.tag == "testsuites":
        target = config.get("xunit_target")
    elif xml_root.tag == "requirements":
        target = config.get("requirement_target")
    else:
        raise Dump2PolarionException("Failed to submit to Polarion - submit target not found")
    return target


def _get_queue_url(xml_root, config):
    if xml_root.tag == "testcases":
        target = config.get("testcase_queue")
    elif xml_root.tag == "testsuites":
        target = config.get("xunit_queue")
    elif xml_root.tag == "requirements":
        target = config.get("requirement_queue")
    else:
        return None
    return target


def _get_log_url(xml_root, config):
    if xml_root.tag == "testcases":
        target = config.get("testcase_log")
    elif xml_root.tag == "testsuites":
        target = config.get("xunit_log")
    elif xml_root.tag == "requirements":
        target = config.get("requirement_log")
    else:
        return None
    return target


def response2dict(response):
    """Returns dict of the response."""
    try:
        return response.json()
    # pylint: disable=broad-except
    except Exception:
        return None


def get_job_ids(response):
    """Returns job IDs of the import."""
    if not isinstance(response, dict):
        response = response2dict(response)
    if response:
        return response["files"]["results.xml"]["job-ids"]
    return None


def get_error_message(response):
    """Returns job IDs of the import."""
    if not isinstance(response, dict):
        response = response2dict(response)
    if response:
        return response["files"]["results.xml"].get("error-message")
    return None


def validate_response(response, submit_target):
    """Checks that the response is valid and import succeeded."""
    if response is None:
        logger.error("Failed to submit to %s", submit_target)
        return False

    if not response:
        logger.error("HTTP status %d: failed to submit to %s", response.status_code, submit_target)
        return False

    parsed_response = response2dict(response)

    if not parsed_response:
        logger.error("Submit to %s failed, invalid response received", submit_target)
        return False

    error_message = get_error_message(parsed_response)
    if error_message:
        logger.error("Submit to %s failed with error", submit_target)
        logger.debug("Error message: %s", error_message)
        return False

    job_ids = get_job_ids(parsed_response)
    if not job_ids or job_ids == [0]:
        logger.error("Submit to %s failed to get job id", submit_target)
        return False

    logger.info("Results received by the Importer (HTTP status %d)", response.status_code)
    logger.info("Job IDs: %s", job_ids)

    return response


def _get_credentials(config, **kwargs):
    login = kwargs.get("user") or config.get("username") or os.environ.get("POLARION_USERNAME")
    pwd = kwargs.get("password") or config.get("password") or os.environ.get("POLARION_PASSWORD")

    if not all([login, pwd]):
        raise Dump2PolarionException("Failed to submit to Polarion - missing credentials")

    return (login, pwd)


# pylint: disable=too-many-arguments
def submit(
    xml_str=None, xml_file=None, xml_root=None, config=None, session=None, dry_run=None, **kwargs
):
    """Submits data to the Polarion Importer."""
    try:
        config = config or configuration.get_config()
        xml_root = _get_xml_root(xml_root, xml_str, xml_file)
        credentials = _get_credentials(config, **kwargs)
        submit_target = _get_submit_target(xml_root, config)
        properties.xunit_fill_testrun_id(xml_root, kwargs.get("testrun_id"))
        if dry_run is not None:
            properties.set_dry_run(xml_root, dry_run)
        xml_input = utils.etree_to_string(xml_root)
        session = session or utils.get_session(credentials, config)
    except Dump2PolarionException as err:
        logger.error(err)
        return None

    logger.info("Submitting data to %s", submit_target)
    files = {"file": ("results.xml", xml_input)}
    try:
        response = session.post(submit_target, files=files)
    # pylint: disable=broad-except
    except Exception as err:
        logger.error(err)
        response = None

    return validate_response(response, submit_target)


# pylint: disable=too-many-arguments
def submit_and_verify(
    xml_str=None, xml_file=None, xml_root=None, config=None, session=None, dry_run=None, **kwargs
):
    """Submits data to the Polarion Importer and checks that it was imported."""
    try:
        config = config or configuration.get_config()
        xml_root = _get_xml_root(xml_root, xml_str, xml_file)
        credentials = _get_credentials(config, **kwargs)
        queue_url = _get_queue_url(xml_root, config)
        log_url = _get_log_url(xml_root, config)
        session = session or utils.get_session(credentials, config)
    except Dump2PolarionException as err:
        logger.error(err)
        return None

    response = submit(xml_root=xml_root, config=config, session=session, dry_run=dry_run, **kwargs)
    if not response:
        return False

    job_ids = get_job_ids(response)
    if not kwargs.get("no_verify") and job_ids:
        response = verify_submit(
            session,
            queue_url,
            log_url,
            job_ids,
            timeout=kwargs.get("verify_timeout"),
            log_file=kwargs.get("log_file"),
        )

    return bool(response)



# -*- coding: utf-8 -*-
"""
Exceptions.
"""


class Dump2PolarionException(Exception):
    """dump2polarion exception."""


class NothingToDoException(Dump2PolarionException):
    """nothing to do exception."""



# -*- coding: utf-8 -*-
"""
Creates a Requirement XML file for submitting to the Polarion Importer.

Example of input requirements_data:
requirements_data = [
    {
        "title": "requirement_complete",
        "description": "Complete Requirement",
        "approver-ids": "mkourim:approved",
        "assignee-id": "mkourim",
        "category-ids": "category_id1, category_id2",
        "due-date": "2018-09-30",
        "planned-in-ids": "planned_id1, planned_id2",
        "initial-estimate": "1/4h",
        "priority-id": "high",
        "severity-id": "should_have",
        "status-id": "status_id",
        "reqtype": "functional",
    },
    {
        "title": "requirement_minimal",
    },
]
"""

from __future__ import absolute_import, unicode_literals

import datetime
from collections import OrderedDict

import six
from lxml import etree

from dump2polarion import transform, utils
from dump2polarion.exceptions import Dump2PolarionException, NothingToDoException


class RequirementExport(object):
    """Exports requirements data into XML representation."""

    REQ_DATA = OrderedDict(
        (
            ("approver-ids", None),
            ("assignee-id", None),
            ("category-ids", None),
            ("due-date", None),
            ("initial-estimate", None),
            ("planned-in-ids", None),
            ("priority-id", "high"),
            ("severity-id", "should_have"),
            ("status-id", None),
        )
    )
    CUSTOM_FIELDS = OrderedDict((("reqtype", "functional"),))

    def __init__(self, requirements_data, config, transform_func=None):
        self.requirements_data = requirements_data
        self.config = config
        self._lookup_prop = ""
        self._transform_func = transform_func or transform.get_requirements_transform(config)

    def _transform_result(self, result):
        """Calls transform function on result."""
        if self._transform_func:
            result = self._transform_func(result)
        return result or None

    def _top_element(self):
        """Returns top XML element."""
        attrs = {"project-id": self.config["polarion-project-id"]}
        document_relative_path = self.config.get("requirements-document-relative-path")
        if document_relative_path:
            attrs["document-relative-path"] = document_relative_path
        top = etree.Element("requirements", attrs)
        return top

    def _properties_element(self, parent_element):
        """Returns properties XML element."""
        requirements_properties = etree.SubElement(parent_element, "properties")

        req_properties_conf = self.config.get("requirements_import_properties") or {}
        for name, value in sorted(six.iteritems(req_properties_conf)):
            if name == "lookup-method":
                lookup_prop = str(value).lower()
                if lookup_prop not in ("id", "name"):
                    raise Dump2PolarionException(
                        "Invalid value '{}' for the 'lookup-method' property".format(str(value))
                    )
                self._lookup_prop = lookup_prop
            else:
                etree.SubElement(
                    requirements_properties, "property", {"name": name, "value": str(value)}
                )

        return requirements_properties

    def _fill_lookup_prop(self, requirements_properties):
        """Fills the polarion-lookup-method property."""
        if not self._lookup_prop:
            raise Dump2PolarionException("Failed to set the 'polarion-lookup-method' property")

        etree.SubElement(
            requirements_properties,
            "property",
            {"name": "lookup-method", "value": self._lookup_prop},
        )

    def _check_lookup_prop(self, req_id):
        """Checks that selected lookup property can be used for this testcase."""
        if self._lookup_prop:
            if not req_id and self._lookup_prop == "id":
                return False
        else:
            if req_id:
                self._lookup_prop = "id"
            else:
                self._lookup_prop = "name"
        return True

    def _classify_data(self, req_data):
        attrs = OrderedDict()
        custom_fields = OrderedDict()

        for key, value in six.iteritems(req_data):
            if not value:
                continue
            if key in self.REQ_DATA:
                attrs[key] = value
            elif key in self.CUSTOM_FIELDS:
                custom_fields[key] = value

        return attrs, custom_fields

    def _fill_defaults(self, attrs, custom_fields):
        for key, value in six.iteritems(self.REQ_DATA):
            if value and not attrs.get(key):
                attrs[key] = value
        for key, value in six.iteritems(self.CUSTOM_FIELDS):
            if value and not custom_fields.get(key):
                custom_fields[key] = value
        return attrs, custom_fields

    @staticmethod
    def _fill_custom_fields(parent, custom_fields):
        if not custom_fields:
            return

        custom_fields_el = etree.SubElement(parent, "custom-fields")
        for field, content in six.iteritems(custom_fields):
            etree.SubElement(custom_fields_el, "custom-field", {"id": field, "content": content})

    def _requirement_element(self, parent_element, req_data):
        """Adds requirement XML element."""
        req_data = self._transform_result(req_data)
        if not req_data:
            return

        title = req_data.get("title")
        if not title:
            return
        req_id = req_data.get("id")

        if not self._check_lookup_prop(req_id):
            return

        attrs, custom_fields = self._classify_data(req_data)
        attrs, custom_fields = self._fill_defaults(attrs, custom_fields)

        requirement = etree.SubElement(parent_element, "requirement", attrs)

        title_el = etree.SubElement(requirement, "title")
        title_el.text = title

        description = req_data.get("description")
        if description:
            description_el = etree.SubElement(requirement, "description")
            description_el.text = description

        self._fill_custom_fields(requirement, custom_fields)

    def _fill_requirements(self, parent_element):
        if not self.requirements_data:
            raise NothingToDoException("Nothing to export")
        for req_data in self.requirements_data:
            self._requirement_element(parent_element, req_data)

    def export(self):
        """Returns requirements XML."""
        top = self._top_element()
        properties = self._properties_element(top)
        self._fill_requirements(top)
        self._fill_lookup_prop(properties)
        return utils.prettify_xml(top)

    @staticmethod
    def write_xml(xml, output_file=None):
        """Outputs the XML content into a file."""
        gen_filename = "requirements-{:%Y%m%d%H%M%S}.xml".format(datetime.datetime.now())
        utils.write_xml(xml, output_loc=output_file, filename=gen_filename)



# -*- coding: utf-8 -*-
"""
Creates a Testcase XML file for submitting to the Polarion Importer.

Example of input testcases_data:
testcases_data = [
    {
        "id": "ITEM01",
        "title": "test_manual",
        "description": "Manual tests with all supported fields.",
        "approver-ids": "mkourim:approved",
        "assignee-id": "mkourim",
        "due-date": "2018-09-30",
        "initial-estimate": "1/4h",
        "caseautomation": "manualonly",
        "caseimportance": "high",
        "caselevel": "component",
        "caseposneg": "positive",
        "testtype": "functional",
        "subtype1": "-",
        "subtype2": "-",
        "upstream": "yes",
        "tags": "tag1, tag2",
        "setup": "Do this first",
        "teardown": "Clean up",
        "automation_script": "https://gitlab.com/foo",
        "testSteps": ["step1", "step2"],
        "expectedResults": ["result1", "result2"],
        "linked-items": "ITEM01",
    },
    {
        "title": "test_minimal_param",
        "params": ["param1", "param2"],
        "linked-items": [{"id": "ITEM01", "role": "derived_from"}],
    },
]
"""

from __future__ import absolute_import, unicode_literals

import datetime

from collections import OrderedDict

import six

from lxml import etree

from dump2polarion import transform, utils
from dump2polarion.exceptions import Dump2PolarionException, NothingToDoException


class TestcaseExport(object):
    """Exports testcases data into XML representation."""

    TESTCASE_DATA = OrderedDict(
        (
            ("approver-ids", None),
            ("assignee-id", None),
            ("due-date", None),
            ("id", None),
            ("initial-estimate", None),
        )
    )
    CUSTOM_FIELDS = OrderedDict(
        (
            ("automation_script", None),
            ("caseautomation", "automated"),
            ("caseimportance", "high"),
            ("caselevel", "component"),
            ("caseposneg", "positive"),
            ("setup", None),
            ("subtype1", "-"),
            ("subtype2", "-"),
            ("tags", None),
            ("teardown", None),
            ("testtype", "functional"),
            ("upstream", None),
        )
    )

    def __init__(self, testcases_data, config, transform_func=None):
        self.testcases_data = testcases_data
        self.config = config
        self._lookup_prop = ""
        self._transform_func = transform_func or transform.get_testcases_transform(config)

    def _transform_result(self, result):
        """Calls transform function on result."""
        if self._transform_func:
            result = self._transform_func(result)
        return result or None

    def _top_element(self):
        """Returns top XML element."""
        attrs = {"project-id": self.config["polarion-project-id"]}
        top = etree.Element("testcases", attrs)
        return top

    def _properties_element(self, parent_element):
        """Returns properties XML element."""
        testcases_properties = etree.SubElement(parent_element, "properties")

        testcases_properties_conf = self.config.get("testcase_import_properties") or {}
        for name, value in sorted(six.iteritems(testcases_properties_conf)):
            if name == "lookup-method":
                lookup_prop = str(value).lower()
                if lookup_prop not in ("id", "name", "custom"):
                    raise Dump2PolarionException(
                        "Invalid value '{}' for the 'lookup-method' property".format(str(value))
                    )
                self._lookup_prop = lookup_prop
            else:
                etree.SubElement(
                    testcases_properties, "property", {"name": name, "value": str(value)}
                )

        return testcases_properties

    def _fill_lookup_prop(self, testcases_properties):
        """Fills the polarion-lookup-method property."""
        if not self._lookup_prop:
            raise Dump2PolarionException("Failed to set the 'polarion-lookup-method' property")

        etree.SubElement(
            testcases_properties, "property", {"name": "lookup-method", "value": self._lookup_prop}
        )

    def _check_lookup_prop(self, testcase_id, testcase_title):
        """Checks that selected lookup property can be used for this testcase."""
        if self._lookup_prop:
            if not testcase_id and self._lookup_prop != "name":
                return None
            if not testcase_title and self._lookup_prop == "name":
                return None
        else:
            if testcase_id:
                self._lookup_prop = "id"
            elif testcase_title:
                self._lookup_prop = "name"
        return True

    def _classify_data(self, testcase_data):
        attrs = OrderedDict()
        custom_fields = OrderedDict()

        for key, value in six.iteritems(testcase_data):
            if not value:
                continue
            if key in self.TESTCASE_DATA:
                attrs[key] = value
            elif key in self.CUSTOM_FIELDS:
                custom_fields[key] = value

        return attrs, custom_fields

    def _fill_defaults(self, attrs, custom_fields):
        for key, value in six.iteritems(self.TESTCASE_DATA):
            if value and not attrs.get(key):
                attrs[key] = value
        for key, value in six.iteritems(self.CUSTOM_FIELDS):
            if value and not custom_fields.get(key):
                custom_fields[key] = value
        return attrs, custom_fields

    @staticmethod
    def _add_test_steps(parent, testcase_data):
        steps = testcase_data.get("testSteps")
        results = testcase_data.get("expectedResults")
        params = testcase_data.get("params") or ()
        test_steps = etree.SubElement(parent, "test-steps")

        if steps and testcase_data["caseautomation"] != "automated":
            for index, step in enumerate(steps):
                test_step = etree.SubElement(test_steps, "test-step")
                test_step_col = etree.SubElement(test_step, "test-step-column", id="step")
                test_step_col.text = step

                test_res_col = etree.SubElement(test_step, "test-step-column", id="expectedResult")
                try:
                    test_res_col.text = results[index]
                except IndexError:
                    test_res_col.text = ""
        else:
            test_step = etree.SubElement(test_steps, "test-step")
            test_step_col = etree.SubElement(test_step, "test-step-column", id="step")
            for param in params:
                param_el = etree.Element("parameter", name=param, scope="local")
                test_step_col.append(param_el)

    @staticmethod
    def _add_linked_items(parent, testcase_data):
        linked_items = testcase_data.get("linked-items") or testcase_data.get("linked-work-items")
        if not linked_items:
            return
        if isinstance(linked_items, (dict, six.string_types)):
            linked_items = [linked_items]

        linked_work_items = etree.SubElement(parent, "linked-work-items")
        for work_item in linked_items:
            if isinstance(work_item, six.string_types):
                work_item_id = work_item
                work_item_role = "verifies"
            else:
                work_item_id = work_item.get("id")
                work_item_role = work_item.get("role") or "verifies"

            if not work_item_id:
                continue

            work_item_el = etree.SubElement(linked_work_items, "linked-work-item")
            work_item_el.attrib["workitem-id"] = work_item_id
            work_item_el.attrib["role-id"] = work_item_role

            lookup_method = testcase_data.get("linked-items-lookup-method")
            if lookup_method in ("name", "id"):
                work_item_el.attrib["lookup-method"] = lookup_method

    @staticmethod
    def _fill_custom_fields(parent, custom_fields):
        if not custom_fields:
            return

        custom_fields_el = etree.SubElement(parent, "custom-fields")
        for field, content in six.iteritems(custom_fields):
            etree.SubElement(custom_fields_el, "custom-field", {"id": field, "content": content})

    def _testcase_element(self, parent_element, testcase_data):
        """Adds testcase XML element."""
        testcase_data = self._transform_result(testcase_data)
        if not testcase_data:
            return

        title = testcase_data.get("title")
        testcase_id = testcase_data.get("id")
        if not (title or testcase_id):
            return

        if not self._check_lookup_prop(testcase_id, title):
            return

        attrs, custom_fields = self._classify_data(testcase_data)
        attrs, custom_fields = self._fill_defaults(attrs, custom_fields)

        testcase = etree.SubElement(parent_element, "testcase", attrs)

        title_el = etree.SubElement(testcase, "title")
        title_el.text = title

        description = testcase_data.get("description")
        if description:
            description_el = etree.SubElement(testcase, "description")
            description_el.text = description

        self._add_test_steps(testcase, testcase_data)
        self._fill_custom_fields(testcase, custom_fields)
        self._add_linked_items(testcase, testcase_data)

    def _fill_testcases(self, parent_element):
        if not self.testcases_data:
            raise NothingToDoException("Nothing to export")
        for testcase_data in self.testcases_data:
            self._testcase_element(parent_element, testcase_data)

    def export(self):
        """Returns testcases XML."""
        top = self._top_element()
        properties = self._properties_element(top)
        self._fill_testcases(top)
        self._fill_lookup_prop(properties)
        return utils.prettify_xml(top)

    @staticmethod
    def write_xml(xml, output_file=None):
        """Outputs the XML content into a file."""
        gen_filename = "testcases-{:%Y%m%d%H%M%S}.xml".format(datetime.datetime.now())
        utils.write_xml(xml, output_loc=output_file, filename=gen_filename)



# -*- coding: utf-8 -*-
"""CSV Unicode Reader."""

import codecs
import csv


def get_csv_reader(csvfile, dialect=csv.excel, encoding="utf-8", **kwds):
    """Returns csv reader."""
    try:
        # pylint: disable=pointless-statement
        unicode
        return UnicodeReader(csvfile, dialect=dialect, encoding=encoding, **kwds)
    except NameError:
        return csv.reader(csvfile, dialect=dialect, **kwds)


# pylint: disable=too-few-public-methods,non-iterator-returned
class UTF8Recoder(object):
    """Iterator that reads an encoded stream and reencodes the input to UTF-8."""

    def __init__(self, csvfile, encoding):
        self.reader = codecs.getreader(encoding)(csvfile)

    def __iter__(self):
        return self

    def next(self):
        """Returns the next row of the reader’s iterable object."""
        return self.reader.next().encode("utf-8")


# pylint: disable=too-few-public-methods
class UnicodeReader(object):
    """A CSV reader which will iterate over lines in the CSV file.

    The CSV file is encoded in the given encoding.
    """

    def __init__(self, csvfile, dialect=csv.excel, encoding="utf-8", **kwds):
        csvfile = UTF8Recoder(csvfile, encoding)
        self.reader = csv.reader(csvfile, dialect=dialect, **kwds)

    def next(self):
        """Returns the next row of the reader’s iterable object."""
        row = self.reader.next()
        return [s.decode("utf-8") for s in row]

    def __iter__(self):
        return self



# -*- coding: utf-8 -*-
"""
Allowed values for verdicts.
"""

from __future__ import unicode_literals


# pylint: disable=too-few-public-methods
class Verdicts(object):
    """Valid verdicts."""

    PASS = ("passed", "pass")
    FAIL = ("failed", "fail")
    SKIP = ("skipped", "skip", "blocked")
    WAIT = ("null", "wait", "waiting")



# -*- coding: utf-8 -*-
"""
Dump testcases results from a CSV, SQLite, junit or Ostriz input file to XUnit file.
Submit XUnit, Testcases or Requirements XML to the Polarion Importers.
"""

from __future__ import absolute_import, unicode_literals

import argparse
import datetime
import io
import logging
import os

import dump2polarion
from dump2polarion import dbtools, utils
from dump2polarion.exceptions import Dump2PolarionException, NothingToDoException

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def get_args(args=None):
    """Get command line arguments."""
    parser = argparse.ArgumentParser(description="dump2polarion")
    parser.add_argument(
        "-i",
        "--input_file",
        required=True,
        help="Path to CSV, SQLite or JUnit reports file or importers XML file",
    )
    parser.add_argument(
        "-o", "--output_file", help="Where to save the XML output file (default: not saved)"
    )
    parser.add_argument("-t", "--testrun-id", help="Polarion test run id")
    parser.add_argument("-c", "--config-file", help="Path to config YAML")
    parser.add_argument(
        "-n", "--no-submit", action="store_true", help="Don't submit results to Polarion"
    )
    parser.add_argument("--user", help="Username to use to submit results to Polarion")
    parser.add_argument("--password", help="Password to use to submit results to Polarion")
    parser.add_argument("-f", "--force", action="store_true", help="Don't validate test run id")
    parser.add_argument("--dry-run", action="store_true", help="Dry run, don't update anything")
    parser.add_argument("--no-verify", action="store_true", help="Don't verify results submission")
    parser.add_argument(
        "--verify-timeout",
        type=int,
        default=300,
        metavar="SEC",
        help="How long to wait (in seconds) for verification of results submission"
        " (default: %(default)s)",
    )
    parser.add_argument(
        "--job-log",
        help="Where to save the log file produced by the Importer" " (default: not saved)",
    )
    parser.add_argument("--log-level", help="Set logging to specified level")
    return parser.parse_args(args)


def get_submit_args(args):
    """Gets arguments for the ``submit_and_verify`` method."""
    submit_args = dict(
        testrun_id=args.testrun_id,
        user=args.user,
        password=args.password,
        no_verify=args.no_verify,
        verify_timeout=args.verify_timeout,
        log_file=args.job_log,
        dry_run=args.dry_run,
    )
    return {k: v for k, v in submit_args.items() if v is not None}


def get_testrun_id(args, testrun_id):
    """Returns testrun id."""
    if args.testrun_id and testrun_id and not args.force and testrun_id != args.testrun_id:
        raise Dump2PolarionException(
            "The test run id '{}' found in exported data doesn't match '{}'. "
            "If you really want to proceed, add '-f'.".format(testrun_id, args.testrun_id)
        )

    found_testrun_id = args.testrun_id or testrun_id
    if not found_testrun_id:
        raise Dump2PolarionException(
            "The testrun id was not specified on command line and not found in the input data."
        )

    return found_testrun_id


def submit_if_ready(args, submit_args, config):
    """Submits the input XML file if it's already in the expected format."""
    __, ext = os.path.splitext(args.input_file)
    if ext.lower() != ".xml":
        return None

    with io.open(args.input_file, encoding="utf-8") as input_file:
        xml = input_file.read(1024)

    if not ("<testsuites" in xml or "<testcases" in xml or "<requirements" in xml):
        return None

    if args.no_submit:
        logger.info("Nothing to do")
        return 0

    # expect importer xml and just submit it
    response = dump2polarion.submit_and_verify(
        xml_file=args.input_file, config=config, **submit_args
    )
    return 0 if response else 2


def main(args=None, transform_func=None):
    """Main function for cli."""
    args = get_args(args)
    submit_args = get_submit_args(args)

    utils.init_log(args.log_level)

    try:
        config = dump2polarion.get_config(args.config_file)
    except Dump2PolarionException as err:
        logger.fatal(err)
        return 1

    submit_outcome = submit_if_ready(args, submit_args, config)
    if submit_outcome is not None:
        # submitted, nothing more to do
        return submit_outcome

    import_time = datetime.datetime.utcnow()

    try:
        records = dump2polarion.do_import(args.input_file, older_than=import_time)
        testrun_id = get_testrun_id(args, records.testrun)
        exporter = dump2polarion.XunitExport(
            testrun_id, records, config, transform_func=transform_func
        )
        output = exporter.export()
    except NothingToDoException as info:
        logger.info(info)
        return 0
    except (EnvironmentError, Dump2PolarionException) as err:
        logger.fatal(err)
        return 1

    if args.output_file or args.no_submit:
        # when no output file is specified, the 'testrun_TESTRUN_ID-TIMESTAMP'
        # file will be created in current directory
        exporter.write_xml(output, args.output_file)

    if not args.no_submit:
        response = dump2polarion.submit_and_verify(output, config=config, **submit_args)

        __, ext = os.path.splitext(args.input_file)
        if ext.lower() in dbtools.SQLITE_EXT and response:
            dbtools.mark_exported_sqlite(args.input_file, import_time)

        return 0 if response else 2

    return 0



# -*- coding: utf-8 -*-
"""
Helper functions for handling data in pytest junit format.
"""

from __future__ import absolute_import, unicode_literals

import os
from collections import OrderedDict

import six
from lxml import etree

from dump2polarion import xunit_exporter
from dump2polarion.exceptions import Dump2PolarionException

_PARAMETER_PREFIX = "polarion-parameter-"


def _get_xml_root(junit_file):
    try:
        tree = etree.parse(os.path.expanduser(junit_file))
    except Exception as err:
        raise Dump2PolarionException("Failed to parse XML file '{}': {}".format(junit_file, err))

    return tree.getroot()


def _parse_testcase_record(testcase_record):
    """Parses testcase record and returns it's info."""
    verdict = None
    verdict_found = False
    comment = ""
    properties = {}
    for element in testcase_record:
        if not verdict_found:
            if element.tag == "error":
                verdict = "failed"
                comment = element.get("message")
                # continue to see if there's more telling verdict for this record
            elif element.tag == "failure":
                verdict = "failed"
                comment = element.get("message")
                verdict_found = True
            elif element.tag == "skipped":
                verdict = "skipped"
                comment = element.get("message")
                verdict_found = True
        if element.tag == "properties":
            for prop in element:
                properties[prop.get("name")] = prop.get("value")
    if not verdict:
        verdict = "passed"

    return verdict, comment, properties


def _extract_parameters_from_properties(properties):
    """Extracts parameters from properties."""
    new_properties = {}
    parameters = []
    for key, value in six.iteritems(properties):
        if key.startswith(_PARAMETER_PREFIX):
            parameters.append((key.replace(_PARAMETER_PREFIX, ""), value))
        else:
            new_properties[key] = value

    return new_properties, sorted(parameters)


# pylint: disable=unused-argument
def import_junit(junit_file, **kwargs):
    """Reads the content of the junit-results file produced by pytest and returns imported data."""
    xml_root = _get_xml_root(junit_file)

    results = []
    for test_data in xml_root:
        if test_data.tag != "testcase":
            continue

        verdict, comment, properties = _parse_testcase_record(test_data)
        properties, parameters = _extract_parameters_from_properties(properties)

        title = test_data.get("name")
        classname = test_data.get("classname")
        time = test_data.get("time", 0)
        filepath = test_data.get("file")

        data = [
            ("title", title),
            ("classname", classname),
            ("verdict", verdict),
            ("comment", comment),
            ("time", time),
            ("file", filepath),
        ]
        for key in sorted(properties):
            data.append((key, properties[key]))
        if parameters:
            data.append(("params", OrderedDict(parameters)))

        results.append(OrderedDict(data))

    return xunit_exporter.ImportedData(results=results, testrun=None)
